var fs = require('fs')
  , packageInfo = require('./package.json')
  , program = require('commander')
  , parser = require('./core/parser');

program
  .version(packageInfo.version)
  .description('The tool convert excel to protobuf serializable file and output .proto & .cs files')
  .usage('[options]')
  .option('-i, --input [type]', 'Input directory for excel files (required)', checkPathExists)
  .option('-o, --output [type]', 'Output directory (required)', checkPathExists)
  .option('-ns --ns [type]', 'Proto file namespace')
  .option('-p --proto [type]', 'Proto file name')
  .option('-cs --cs [type]', 'CSharp file name')
  .parse(process.argv);

if(!program.input || !program.output) {
  console.log('Some options are required, or using option -h look help.'.yellow);
  return;
}

let outputDir = `${program.output}/output`;
let outputTmpDir = `${outputDir}/tmp`;
let outputSerialDir = `${outputDir}/serialized`;

let options = {
  ns: program.ns ? program.ns : 'ProtoMessages',
  protoFilePath: program.p ? `${outputDir}/${program.p}` : `${outputDir}/messages.proto`,
  csFilePath: program.cs ? `${outputDir}/${program.cs}` : `${outputDir}/Messages.cs`,
};

// Create output folders
if (fs.existsSync(outputDir)) {
  deleteFolderRecursive(outputDir);
}
fs.mkdirSync(outputDir);
if (!fs.existsSync(outputSerialDir)) {
  fs.mkdirSync(outputSerialDir);
}
if (!fs.existsSync(outputTmpDir)) {
  fs.mkdirSync(outputTmpDir);
}

parserFiles(xlsies => {
  let content = `package ${options.ns};\n`;
  xlsies.forEach(file => {
    content += parser.parserSync(file, options);
  });
  parser.generateCSharp(content, options.protoFilePath, options.csFilePath);
});

function parserFiles (cb){
  let xlsies = [];
  fs.readdir(program.input, (err, files) => {
    files.forEach(file => {
      if (file.endsWith('.xlsx')) {
        xlsies.push(file)
      }
    });
    cb(xlsies)
  });
}

function deleteFolderRecursive(path) {
  if (fs.existsSync(path)) {
    fs.readdirSync(path).forEach(function (file, index) {
      var curPath = path + "/" + file;
      if (fs.lstatSync(curPath).isDirectory()) { // recurse
        deleteFolderRecursive(curPath);
      } else { // delete file
        fs.unlinkSync(curPath);
      }
    });
    fs.rmdirSync(path);
  }
};

function checkPathExists(filepath){
  if(!fs.existsSync(filepath)) console.log((filepath + ' path dose not exists').red);
  return filepath;
}
