var xlsx = require('node-xlsx').default
  , colors = require('colors')
  , fs = require('fs')
  , path = require('path')
  , mustache = require('mustache')
  , protobuf = require('protocol-buffers')
  , exec = require('child_process').execFile
  , _ = require('lodash');

var protoGenExe = __dirname + '/proto-gen/protogen.exe';

exports.parserSync = function(xlsxFilePath, options) {
  let viewData = GetXlsxDataSync(xlsxFilePath, options);
  let tmpProtoFilePath = path.dirname(options.protoFilePath) + '/tmp/' + (new Date().getTime()) + '.proto';
  // generate temp proto file
  let tpl = fs.readFileSync(__dirname + '/template.mustache', 'utf-8');
  let view = mustache.render(tpl, viewData);
  fs.writeFileSync(tmpProtoFilePath, view);

  // serialize
  let messages = protobuf(fs.readFileSync(tmpProtoFilePath));
  // one view data one xlsxfile
  viewData.messages.forEach(msg => {
    let encodeCmd = `messages.${msg.name}s.encode({datas: ${JSON.stringify(GetSheetDataArray(msg.datas))} })`;
    fs.writeFileSync(`${path.dirname(options.protoFilePath)}/serialized/${msg.name}s.bytes`, eval(encodeCmd))
    console.log(`${msg.name}s.bytes generated.`.green);
  });
  console.log('---------------------------------------'.cyan);
  let lines = view.split('\n');
  lines.splice(0, 1);
  return lines.join('\n');
};

exports.generateCSharp = function(content, protoFilePath, csFilePath){
  fs.writeFileSync(protoFilePath, content);
  console.log(`${path.basename(protoFilePath)} generated`.green);
  exec(protoGenExe, [`-i:${protoFilePath}`, `-o:${csFilePath}`], (err, result) => {
    if (err) {
      console.log(err);
    } else {
      console.log(`${path.basename(csFilePath)} generated.`.green);
    }
  });
};

function GetSheetDataArray(data){
  return _.map(data, (arr) => {
    return _.reduce(arr, (result, item) => {
      return _.assign(result, item)
    }, {});
  });
}

function GetXlsxDataSync(xlsxFilePath, options){
  console.log(`Parsing ${path.basename(xlsxFilePath)}`.green);
  let viewData = {
    namespace: options.ns,
    messages: []
  };
  let sheets = xlsx.parse(xlsxFilePath);

  sheets.forEach((st) => {
    let names = st.name.split('_');
    if (names.length !== 2) {
      return console.log(`Please check sheet name: ${st.name}`.red);
    }
    let message = {
      name: names[1],
      fields: [],
      datas: []
    };

    let colCount = 0;
    st.data[0].forEach(col => {
      if (!col.startsWith('#')) {
        colCount++;
      }
    });

    // First row is fieldName, second row is fieldType
    for (let i = 0; i < colCount; i++) {
      let type = st.data[1][i].startsWith('required') ? st.data[1][i] : 'optional ' + st.data[1][i];
      message.fields.push({type: type, name: st.data[0][i], number: (i+1)});
    }
    for(let i = 3; i < st.data.length; i++) { // exclude 3rd row
      let rowData = [];
      for (let j = 0; j < colCount; j++) {
        let dt = {};
        dt[message.fields[j].name] = st.data[i][j];
        rowData.push(dt);
      }
      message.datas.push(rowData);
    }
    viewData.messages.push(message);
  });

  return viewData;
}
